from typing import Optional

import numpy as np
from scipy.linalg import eig, eigh, solve
from scipy.sparse.linalg import eigsh
from sklearn.utils.extmath import randomized_svd

from kooplearn._src.linalg import eigh_rank_reveal, spd_neg_pow, weighted_norm
from kooplearn._src.utils import fuzzy_parse_complex, topk


def fit_reduced_rank_regression(
    C_X: np.ndarray,  # Input covariance matrix
    C_XY: np.ndarray,  # Cross-covariance matrix
    tikhonov_reg: float,  # Tikhonov regularization parameter, can be 0.0
    rank: int,  # Rank of the estimator
    svd_solver: str = "arnoldi",  # SVD solver to use. Arnoldi is faster for low ranks.
):
    if tikhonov_reg == 0.0:
        return _fit_reduced_rank_regression_noreg(C_X, C_XY, rank, svd_solver)
    else:
        dim = C_X.shape[0]
        reg_input_covariance = C_X + tikhonov_reg * np.identity(dim, dtype=C_X.dtype)
        _crcov = C_XY @ C_XY.T
        if svd_solver == "arnoldi":
            # Adding a small buffer to the Arnoldi-computed eigenvalues.
            values, vectors = eigsh(_crcov, rank + 3, M=reg_input_covariance)
        else:
            values, vectors = eigh(_crcov, reg_input_covariance)

        top_eigs = topk(values, rank)
        vectors = vectors[:, top_eigs.indices]

        _norms = weighted_norm(vectors, reg_input_covariance)
        vectors = vectors @ np.diag(_norms ** (-1.0))
        return vectors


def _fit_reduced_rank_regression_noreg(
    C_X: np.ndarray,  # Input covariance matrix
    C_XY: np.ndarray,  # Cross-covariance matrix
    rank: int,  # Rank of the estimator
    svd_solver: str = "arnoldi",  # SVD solver to use. Arnoldi is faster for low ranks.
):
    rsqrt_C_X = spd_neg_pow(C_X, -0.5)
    B = rsqrt_C_X @ C_XY
    _crcov = B @ B.T
    if svd_solver == "arnoldi":
        # Adding a small buffer to the Arnoldi-computed eigenvalues.
        values, vectors = eigsh(_crcov, rank + 3)
    else:
        values, vectors = eigh(_crcov)
    vectors, _vals, _ = eigh_rank_reveal(values, vectors, rank)
    return rsqrt_C_X @ vectors


def fit_rand_reduced_rank_regression(
    C_X: np.ndarray,  # Input covariance matrix
    C_XY: np.ndarray,  # Cross-covariance matrix
    tikhonov_reg: float,  # Tikhonov regularization parameter
    rank: int,  # Rank of the estimator
    n_oversamples: int,  # Number of oversamples
    iterated_power: int,  # Number of power iterations
    rng_seed: Optional[int] = None,  # Random seed
):
    dim = C_X.shape[0]
    reg_input_covariance = C_X + tikhonov_reg * np.identity(dim, dtype=C_X.dtype)
    _crcov = C_XY @ C_XY.T
    rng = np.random.default_rng(rng_seed)
    sketch = rng.standard_normal(
        size=(reg_input_covariance.shape[0], rank + n_oversamples)
    )

    for _ in range(iterated_power):
        _tmp_sketch = solve(reg_input_covariance, sketch, assume_a="pos")
        sketch = _crcov @ _tmp_sketch
        # TODO add power iteration QR normalization.

    sketch_p = solve(reg_input_covariance, sketch, assume_a="pos")

    F_0 = sketch_p.T @ sketch
    F_1 = sketch_p.T @ _crcov @ sketch_p

    values, vectors = eigh(F_1, F_0)
    _norms = weighted_norm(vectors, F_0)
    vectors = vectors @ np.diag(_norms ** (-1.0))
    return sketch_p @ vectors[:, topk(values, rank).indices]


def fit_principal_component_regression(
    C_X: np.ndarray,  # Input covariance matrix
    tikhonov_reg: float = 0.0,  # Tikhonov regularization parameter, can be 0
    rank: Optional[int] = None,  # Rank of the estimator
    svd_solver: str = "arnoldi",  # SVD solver to use. Arnoldi is faster for low ranks.
):
    dim = C_X.shape[0]
    assert rank <= dim, f"Rank too high. The maximum value for this problem is {dim}"
    reg_input_covariance = C_X + tikhonov_reg * np.identity(dim, dtype=C_X.dtype)
    if svd_solver == "arnoldi":
        values, vectors = eigsh(reg_input_covariance, k=rank, which="LM")
    elif svd_solver == "full":
        values, vectors = eigh(reg_input_covariance)
    else:
        raise ValueError(f"Unknown svd_solver {svd_solver}")

    vectors, _, rsqrt_evals = eigh_rank_reveal(values, vectors, rank)
    return vectors @ np.diag(rsqrt_evals)


def fit_rand_principal_component_regression(
    C_X: np.ndarray,  # Input covariance matrix
    tikhonov_reg: float,  # Tikhonov regularization parameter
    rank: int,  # Rank of the estimator
    n_oversamples: int,  # Number of oversamples
    iterated_power: int,  # Number of power iterations
    rcond: float = 2.2e-16,  # Threshold for the singular values
    rng_seed: Optional[int] = None,  # Random seed
):
    dim = C_X.shape[0]
    assert rank <= dim, f"Rank too high. The maximum value for this problem is {dim}"
    reg_input_covariance = C_X + tikhonov_reg * np.identity(dim, dtype=C_X.dtype)

    vectors, values, _ = randomized_svd(
        reg_input_covariance,
        rank,
        n_oversamples=n_oversamples,
        n_iter=iterated_power,
        random_state=rng_seed,
    )

    vectors, _, rsqrt_evals = eigh_rank_reveal(values, vectors, rank, rcond)
    return vectors @ np.diag(rsqrt_evals)


def predict(
    num_steps: int,  # Number of steps to predict (return the last one)
    U: np.ndarray,  # Projection matrix, as returned by the fit functions defined above
    C_XY: np.ndarray,  # Cross-covariance matrix
    phi_Xin: np.ndarray,  # Feature map evaluated on the initial conditions
    phi_X: np.ndarray,  # Feature map evaluated on the training input data
    obs_train_Y: np.ndarray,  # Observable to be predicted evaluated on the output training data
):
    # G = U U.T C_XY
    # G^n = (U)(U.T C_XY U)^(n-1)(U.T C_XY)
    num_train = phi_X.shape[0]
    phi_Xin_dot_U = phi_Xin @ U
    U_C_XY_U = np.linalg.multi_dot([U.T, C_XY, U])
    U_phi_X_obs_Y = np.linalg.multi_dot([U.T, phi_X.T, obs_train_Y]) * (num_train**-1)
    M = np.linalg.matrix_power(U_C_XY_U, num_steps - 1)
    return np.linalg.multi_dot([phi_Xin_dot_U, M, U_phi_X_obs_Y])


def estimator_eig(
    U: np.ndarray,  # Projection matrix, as returned by the fit functions defined above
    C_XY: np.ndarray,  # Cross-covariance matrix
):
    # Using the trick described in https://arxiv.org/abs/1905.11490
    M = np.linalg.multi_dot([U.T, C_XY, U])
    values, lv, rv = eig(M, left=True, right=True)

    values = fuzzy_parse_complex(values)

    r_perm = np.argsort(values)
    l_perm = np.argsort(values.conj())
    values = values[r_perm]

    # Normalization in RKHS norm
    rv = U @ rv
    rv = rv[:, r_perm]
    rv = rv / np.linalg.norm(rv, axis=0)
    # Biorthogonalization
    lv = np.linalg.multi_dot([C_XY.T, U, lv])
    lv = lv[:, l_perm]
    l_norm = np.sum(lv * rv, axis=0)
    lv = lv / l_norm

    return values, lv, rv


def estimator_modes(
    U: np.ndarray,  # Projection matrix, as returned by the fit functions defined above
    C_XY: np.ndarray,  # Cross-covariance matrix
    phi_X: np.ndarray,  # Feature map evaluated on the training input data
    phi_Xin: np.ndarray,  # Feature map evaluated on the initial conditions
):
    # Using the trick described in https://arxiv.org/abs/1905.11490
    M = np.linalg.multi_dot([U.T, C_XY, U])
    values, lv, rv = eig(M, left=True, right=True)
    values = fuzzy_parse_complex(values)
    r_perm = np.argsort(values)
    l_perm = np.argsort(values.conj())
    # values = values[r_perm]

    # Normalization in RKHS norm
    rv = U @ rv
    rv = rv[:, r_perm]
    rv = rv / np.linalg.norm(rv, axis=0)
    # Biorthogonalization
    lv_full = np.linalg.multi_dot([C_XY.T, U, lv])
    lv_full = lv_full[:, l_perm]
    lv = lv[:, l_perm]
    l_norm = np.sum(lv_full * rv, axis=0)
    lv = lv / l_norm
    r_dim = phi_X.shape[0] ** -1.0

    # Initial conditions
    rv_in = (phi_Xin @ rv).T  # [rank, num_init_conditions]
    # This should be multiplied on the right by the observable evaluated at the output training data
    lv_obs = np.linalg.multi_dot([r_dim * phi_X, U, lv]).T
    return (
        rv_in[:, :, None] * lv_obs[:, None, :]
    )  # [rank, num_init_conditions, num_training_points]


def evaluate_eigenfunction(
    phi_Xin: np.ndarray,  # Feature map evaluated on the initial conditions
    lv_or_rv: np.ndarray,  # Left or right eigenvector, as returned by estimator_eig
):
    return phi_Xin @ lv_or_rv


def svdvals(U, C_XY):
    M = np.linalg.multi_dot([U, U.T, C_XY])
    return np.linalg.svd(M, compute_uv=False)


def estimator_risk(
    cov_Xv: np.ndarray,  # Covariance matrix of the input validation data
    cov_Yv: np.ndarray,  # Covariance matrix of the output validation data
    cov_XYv: np.ndarray,  # Cross-covariance matrix of the validation data
    cov_XY: np.ndarray,  # Cross-covariance matrix of the training data
    U: np.ndarray,  # Projection matrix, as returned by the fit functions defined above
):
    r_Y = np.trace(cov_Yv)
    r_XY = -2 * np.trace(np.linalg.multi_dot([cov_XY.T, U, U.T, cov_XYv]))
    r_X = np.trace(np.linalg.multi_dot([cov_XY.T, U, U.T, cov_Xv, U, U.T, cov_XY]))
    return r_Y + r_XY + r_X



from typing import Callable, Iterable

import numpy as np

from kooplearn.abc import FeatureMap


class IdentityFeatureMap(FeatureMap):
    """Identity feature map returning the input as is."""

    def __call__(self, X: np.ndarray):
        # Standardize shape
        X = np.atleast_2d(X)
        return X


class ConcatenateFeatureMaps(FeatureMap):
    """Concatenate multiple functions into a feature map. The functions should return Numpy arrays which can be concatenated along their last axis.

    Args:
        feature_maps (Iterable of callables): A list of callables which return numpy arrays.
    """

    def __init__(self, feature_maps: Iterable[Callable]):
        super().__init__()
        self.feature_maps = feature_maps

    def __call__(self, X: np.ndarray):
        # Standardize shape
        X = np.asanyarray(X)
        if X.ndim == 0:
            X = X.reshape(1, 1)
        elif X.ndim == 1:
            X = X.reshape(-1, 1)

        X = np.concatenate([fm(X) for fm in self.feature_maps], axis=-1)
        return X



from kooplearn._src.check_deps import check_torch_deps

check_torch_deps()
import torch  # noqa: E402


def sqrtmh(A):
    # Credits to
    """Compute the square root of a Symmetric or Hermitian positive definite matrix or batch of matrices. Credits to  `https://github.com/pytorch/pytorch/issues/25481#issuecomment-1032789228 <https://github.com/pytorch/pytorch/issues/25481#issuecomment-1032789228>`_."""
    L, Q = torch.linalg.eigh(A)
    zero = torch.zeros((), device=L.device, dtype=L.dtype)
    threshold = L.max(-1).values * L.size(-1) * torch.finfo(L.dtype).eps
    L = L.where(L > threshold.unsqueeze(-1), zero)  # zero out small components
    return (Q * L.sqrt().unsqueeze(-2)) @ Q.mH


def vamp_score(cov_X, cov_Y, cov_XY, schatten_norm: int = 2):
    """Variational Approach for learning Markov Processes (VAMP) score by :footcite:t:`Wu2019`.

    Args:
        cov_X (torch.Tensor): Covariance of the initial time steps.
        cov_Y (torch.Tensor): Covariance of the evolved time steps.
        cov_XY (torch.Tensor): Cross-covariance of the initial and evolved time steps.
        schatten_norm (int, optional): Computes the VAMP-p score with ``p = schatten_norm``. Defaults to 2.

    Raises:
        NotImplementedError: If ``schatten_norm`` is not 1 or 2.

    """
    if schatten_norm == 2:
        M = torch.linalg.multi_dot(
            [
                torch.linalg.pinv(cov_X, hermitian=True),
                cov_XY,
                torch.linalg.pinv(cov_Y, hermitian=True),
                cov_XY.T,
            ]
        )
        return torch.trace(M)
    elif schatten_norm == 1:
        sqrt_cov_X = sqrtmh(cov_X)
        sqrt_cov_Y = sqrtmh(cov_Y)
        M = torch.linalg.multi_dot(
            [
                torch.linalg.pinv(sqrt_cov_X, hermitian=True),
                cov_XY,
                torch.linalg.pinv(sqrt_cov_Y, hermitian=True),
            ]
        )
        return torch.linalg.matrix_norm(M, "nuc")
    else:
        raise NotImplementedError(f"Schatten norm {schatten_norm} not implemented")


def relaxed_projection_score(cov_x, cov_y, cov_xy):
    return (torch.linalg.matrix_norm(cov_xy, ord="fro") ** 2) / (
        (
            torch.linalg.matrix_norm(cov_x, ord=2)
            * torch.linalg.matrix_norm(cov_y, ord=2)
        )
    )


def log_fro_metric_deformation_loss(cov):
    eps = torch.finfo(cov.dtype).eps * cov.shape[0]
    # Metric regularization based on Von Neumann entropy
    vals_x = torch.linalg.eigvalsh(cov)
    vals_x = torch.where(vals_x > eps, vals_x, eps)
    loss = torch.mean(-torch.log(vals_x) + vals_x * (vals_x - 1.0))
    return loss



import logging
from typing import Optional
from warnings import warn

import numpy as np
from sklearn.utils import check_array

from kooplearn._src.utils import topk

logger = logging.getLogger("kooplearn")


def spd_neg_pow(
    M: np.ndarray,
    exponent: float = -1.0,
    cutoff: Optional[float] = None,
    strategy: str = "trunc",
) -> np.ndarray:
    """
    Truncated eigenvalue decomposition of A
    """
    if cutoff is None:
        cutoff = 10.0 * M.shape[0] * np.finfo(M.dtype).eps
    w, v = np.linalg.eigh(M)
    if strategy == "trunc":
        sanitized_w = np.where(w <= cutoff, 1.0, w)
        inv_w = np.where(
            w > cutoff, (sanitized_w ** np.abs(exponent)) ** np.sign(exponent), 0.0
        )
        v = np.where(w > cutoff, v, 0.0)
    elif strategy == "tikhonov":
        inv_w = ((w + cutoff) ** np.abs(exponent)) ** np.sign(exponent)
    else:
        raise NotImplementedError(f"Strategy {strategy} not implemented")
    return np.linalg.multi_dot([v, np.diag(inv_w), v.T])


def weighted_norm(A: np.ndarray, M: Optional[np.ndarray] = None):
    r"""Weighted norm of the columns of A.

    Args:
        A (ndarray): 1D or 2D array. If 2D, the columns are treated as vectors.
        M (ndarray or LinearOperator, optional): Weigthing matrix. the norm of the vector :math:`a` is given by
        :math:`\langle a, Ma\rangle`. Defaults to None, corresponding to the Identity matrix. Warning: no checks are
        performed on M being a PSD operator.

    Returns:
        (ndarray or float): If ``A.ndim == 2`` returns 1D array of floats corresponding to the norms of
        the columns of A. Else return a float.
    """
    assert A.ndim <= 2, "'A' must be a vector or a 2D array"
    if M is None:
        norm = np.linalg.norm(A, axis=0)
    else:
        _A = np.dot(M, A)
        _A_T = np.dot(M.T, A)
        norm = np.real(np.sum(0.5 * (np.conj(A) * _A + np.conj(A) * _A_T), axis=0))
    rcond = 10.0 * A.shape[0] * np.finfo(A.dtype).eps
    norm = np.where(norm < rcond, 0.0, norm)
    return np.sqrt(norm)


def weighted_dot_product(A: np.ndarray, B: np.ndarray, M: Optional[np.ndarray] = None):
    """Weighted dot product between the columns of A and B. The output will be equivalent to :math:`A^{*} M B`
    if A and B are 2D arrays.

    Args:
        A, B (ndarray): 1D or 2D arrays.
        M (ndarray or LinearOperator, optional): Weigthing matrix. Defaults to None, corresponding to the
        Identity matrix. Warning: no checks are performed on M being a PSD operator.

    Returns:
        (ndarray or float): The result of :math:`A^{*} M B`.
    """
    assert A.ndim <= 2, "'A' must be a vector or a 2D array"
    assert B.ndim <= 2, "'B' must be a vector or a 2D array"
    A_adj = np.conj(A.T)
    if M is None:
        return np.dot(A_adj, B)
    else:
        _B = np.dot(M, B)
        return np.dot(A_adj, _B)


def _column_pivot(Q, R, k, squared_norms, columns_permutation):
    """
    Helper function to perform column pivoting on the QR decomposition at the k iteration. No checks are performed.
    For internal use only.
    """
    _arg_max = np.argmax(squared_norms[k:])
    j = k + _arg_max
    _in = [k, j]
    _swap = [j, k]
    # Column pivoting
    columns_permutation[_in] = columns_permutation[_swap]
    Q[:, _in] = Q[:, _swap]
    R[:k, _in] = R[:k, _swap]
    squared_norms[_in] = squared_norms[_swap]
    return Q, R, squared_norms, columns_permutation


def modified_QR(
    A: np.ndarray,
    M: Optional[np.ndarray] = None,
    column_pivoting: bool = False,
    rtol: Optional[float] = None,
    verbose: bool = False,
):
    """Modified QR algorithm with column pivoting. Implementation follows the algorithm described in [1].

    Args:
        A (ndarray): 2D array whose columns are vectors to be orthogonalized.
        M (ndarray or LinearOperator, optional): PSD linear operator. If not None, the vectors are orthonormalized with
         respect to the scalar product induced by M. Defaults to None corresponding to Identity matrix.
        column_pivoting (bool, optional): Whether column pivoting is performed. Defaults to False.
        rtol (float, optional): relative tolerance in determining the numerical rank of A. Defaults to 10*A.shape[0]*eps.
        This parameter is used only when ``column_pivoting == True``.
        verbose (bool, optional): Whether to print informations and warnings about the progress of the algorithm.
        Defaults to False.

    Returns:
        tuple: A tuple of the form (Q, R), where Q and R satisfy A = QR. If ``column_pivoting == True``, the permutation
         of the columns of A is returned as well.

    [1] A. Dax: 'A modified Gram–Schmidt algorithm with iterative orthogonalization and column pivoting',
    https://doi.org/10.1016/S0024-3795(00)00022-7.
    """
    A = check_array(A)  # Ensure A is non-empty 2D array containing only finite values.
    num_vecs = A.shape[1]
    effective_rank = num_vecs
    dtype = A.dtype
    Q = np.copy(A)
    R = np.zeros((num_vecs, num_vecs), dtype=dtype)

    if rtol is None:
        rtol = 10.0 * A.shape[0] * np.finfo(A.dtype).eps
    _roundoff = 1e-8  # From reference paper
    _tau = 1e-2  # From reference paper

    columns_permutation = None
    squared_norms = None
    max_norm = None
    norms_error_estimate = None
    if (
        column_pivoting
    ):  # Initialize variables for fast pivoting, without re-evaluation of the norm at each step.
        squared_norms = weighted_norm(Q, M=M) ** 2
        max_norm = np.sqrt(np.max(squared_norms))
        columns_permutation = np.arange(num_vecs)

    for k in range(num_vecs):
        if column_pivoting:
            Q, R, squared_norms, columns_permutation = _column_pivot(
                Q, R, k, squared_norms, columns_permutation
            )
            norms_error_estimate = squared_norms * _roundoff
        if (
            k != 0
        ):  # Reorthogonalization of the column k+1 of A with respect to the previous orthonormal k vectors.
            alpha = weighted_dot_product(
                Q[:, :k], Q[:, k], M=M
            )  # alpha = Q[:,:k].T@M@Q[:,k]
            R[:k, k] += alpha
            Q[:, k] -= np.dot(Q[:, :k], alpha)

        # Numerical rank detection, performed only when column_pivoting == True
        norm_at_iter_k = weighted_norm(Q[:, k], M=M)
        if column_pivoting:
            if norm_at_iter_k < rtol * max_norm:
                effective_rank = k
                if verbose:
                    warn(
                        "Numerical rank of A has been reached with a relative tolerance rtol = {:.2e}. "
                        "Effective rank = {}. Stopping Orthogonalization procedure.".format(
                            rtol, effective_rank
                        )
                    )
                break
                # Normalization of the column k + 1
        R[k, k] = norm_at_iter_k
        Q[:, k] = Q[:, k] / R[k, k]
        # Orthogonalization of the remaining columns with respect to Q[:,k], i.e. the k+1 column of Q.
        if k < num_vecs - 1:
            R[k, k + 1 :] = weighted_dot_product(Q[:, k + 1 :], Q[:, k], M=M)
            Q[:, k + 1 :] -= np.outer(Q[:, k], R[k, k + 1 :])
            # Try fast update of the squared norms, recompute if numerical criteria are not attained.
            if column_pivoting:
                squared_norms[k + 1 :] -= (
                    R[k, k + 1 :] ** 2
                )  # Update norms using Phythagorean Theorem
                update_error_mask = (
                    _tau * squared_norms[k + 1 :] < norms_error_estimate[k + 1 :]
                )  # Check if the error estimate is too large
                if any(update_error_mask):
                    squared_norms[k + 1 :][update_error_mask] = weighted_norm(
                        Q[:, k + 1 :][:, update_error_mask], M=M
                    )  # Recompute the norms if necessary.
    if column_pivoting:
        return (
            Q[:, :effective_rank],
            R[:effective_rank],
            columns_permutation[:effective_rank],
        )
    else:
        return Q[:, :effective_rank], R[:effective_rank]


def eigh_rank_reveal(
    values: np.ndarray,
    vectors: np.ndarray,
    rank: int,  # Desired rank
    rcond: Optional[float] = None,  # Threshold for the singular values
    verbose: bool = True,
):
    if rcond is None:
        rcond = 10.0 * values.shape[0] * np.finfo(values.dtype).eps
    top_vals = topk(values, rank)
    vectors = vectors[:, top_vals.indices]
    values = top_vals.values

    _ftest = values > rcond
    if all(_ftest):
        rsqrt_vals = (np.sqrt(values)) ** -1
    else:
        first_invalid = np.argmax(
            ~_ftest
        )  # In the case of multiple occurrences of the maximum values, the indices corresponding to the first occurrence are returned.
        _first_discarded_val = np.max(np.abs(values[first_invalid:]))
        values = values[_ftest]
        vectors = vectors[:, _ftest]

        if verbose:
            logger.warning(
                f"Warning: Discarted {rank - vectors.shape[1]} dimensions of the {rank} requested due to numerical instability. Consider decreasing the rank. The largest discarded value is: {_first_discarded_val:.3e}."
            )
        # Compute stable sqrt
        rsqrt_vals = (np.sqrt(values)) ** -1
    return vectors, values, rsqrt_vals


def cov(X: np.ndarray, Y: Optional[np.ndarray] = None):
    X = np.atleast_2d(X)
    if X.ndim > 2:
        raise ValueError(f"Input array has more than 2 dimensions ({X.ndim}).")
    rnorm = (X.shape[0]) ** (-0.5)
    X = X * rnorm

    if Y is None:
        c = X.T @ X
    else:
        if X.shape[0] != Y.shape[0]:
            raise ValueError(
                f"Shape mismatch: the covariance between two arrays can be computed only if they have the same initial dimension. Got {X.shape[0]} and {Y.shape[0]}."
            )
        Y = np.atleast_2d(Y)
        if Y.ndim > 2:
            raise ValueError(f"Input array has more than 2 dimensions ({Y.ndim}).")
        Y = Y * rnorm
        c = X.T @ Y
    return c



from __future__ import annotations

import logging
from typing import Callable, Optional

import numpy as np
from sklearn.base import RegressorMixin
from sklearn.gaussian_process.kernels import DotProduct, Kernel

from kooplearn._src.operator_regression import dual
from kooplearn._src.operator_regression.utils import (
    contexts_to_markov_train_states,
    parse_observables,
)
from kooplearn._src.serialization import pickle_load, pickle_save
from kooplearn._src.utils import ShapeError, check_contexts_shape, check_is_fitted
from kooplearn.abc import BaseModel

logger = logging.getLogger("kooplearn")


class KernelDMD(BaseModel, RegressorMixin):
    """
    Kernel Dynamic Mode Decomposition (KernelDMD) Model.
    Implements the KernelDMD estimators approximating the Koopman (deterministic systems) or Transfer (stochastic systems) operator following the approach described in :footcite:t:`Kostic2022`.

    Args:
        kernel (sklearn.gaussian_process.kernels.Kernel): sklearn Kernel object. Defaults to `DotProduct`.
        reduced_rank (bool): If ``True`` initializes the reduced rank estimator introduced in :footcite:t:`Kostic2022`, if ``False`` initializes the classical principal component estimator.
        rank (int): Rank of the estimator. Defaults to 5.
        tikhonov_reg (float): Tikhonov regularization coefficient. ``None`` is equivalent to ``tikhonov_reg = 0``, and internally calls specialized stable algorithms to deal with this specific case.
        svd_solver (str): Solver used to perform the internal SVD calcuations. Currently supported: `full`, uses LAPACK solvers, `arnoldi`, uses ARPACK solvers, `randomized`, uses randomized SVD algorithms as described in :footcite:t:`Turri2023`.
        iterated_power (int): Number of power iterations when using a randomized algorithm (``svd_solver == 'randomized'``).
        n_oversamples (int): Number of oversamples when using a randomized algorithm (``svd_solver == 'randomized'``).
        optimal_sketching (bool): Sketching strategy for the randomized solver. If `True` performs optimal sketching (computaitonally expensive but more accurate).
        rng_seed (int): Random Number Generator seed. Only used when ``svd_solver == 'randomized'``.  Defaults to ``None``, that is no explicit seed is setted.

    .. tip::

        A powerful DMD variation proposed by :footcite:t:`Arbabi2017`, known as Hankel-DMD, evaluates the Koopman/Transfer estimators by stacking consecutive snapshots together in a Hankel matrix. When this model is fitted context windows of length > 2, the lookback window length is automatically set to ``context_len - 1``. Upon fitting, the whole lookback window is flattened and *concatenated* together before evaluating the kernel function, thus realizing an Hankel-KDMD estimator.

    Attributes:
        data_fit : Training data: array of context windows of shape ``(n_samples, context_len, *features_shape)``.
        kernel_X : Kernel matrix evaluated at the initial states, that is ``self.data_fit[:, :self.lookback_len, ...]`. Shape ``(n_samples, n_samples)``
        kernel_Y : Kernel matrix evaluated at the evolved states, that is ``self.data_fit[:, 1:self.lookback_len + 1, ...]``. Shape ``(n_samples, n_samples)``
        kernel_XY : Cross-kernel matrix between initial and evolved states. Shape ``(n_samples, n_samples)``.
        U : Projection matrix of shape (n_samples, rank). The Koopman/Transfer operator is approximated as :math:`k(\cdot, X)U V^T k(\cdot, Y)` (see :footcite:t:`Kostic2022`).
        V : Projection matrix of shape (n_samples, rank). The Koopman/Transfer operator is approximated as :math:`k(\cdot, X)U V^T k(\cdot, Y)` (see :footcite:t:`Kostic2022`).

    """

    def __init__(
        self,
        kernel: Kernel = DotProduct(),
        reduced_rank: bool = True,
        rank: int = 5,
        tikhonov_reg: Optional[float] = None,
        svd_solver: str = "arnoldi",
        iterated_power: int = 1,
        n_oversamples: int = 5,
        optimal_sketching: bool = False,
        rng_seed: Optional[int] = None,
    ):
        # Initial checks
        if svd_solver not in ["full", "arnoldi", "randomized"]:
            raise ValueError(
                "Invalid svd_solver. Allowed values are 'full', 'arnoldi' and 'randomized'."
            )
        if svd_solver == "randomized" and iterated_power < 0:
            raise ValueError("Invalid iterated_power. Must be non-negative.")
        if svd_solver == "randomized" and n_oversamples < 0:
            raise ValueError("Invalid n_oversamples. Must be non-negative.")

        self.rng_seed = rng_seed
        self._kernel = kernel
        if not isinstance(rank, int) or rank < 1:
            raise ValueError("rank must be a positive integer.")

        self.rank = rank
        if tikhonov_reg is None:
            self.tikhonov_reg = 0.0
        else:
            self.tikhonov_reg = tikhonov_reg
        self.svd_solver = svd_solver
        self.iterated_power = iterated_power
        self.n_oversamples = n_oversamples
        self.optimal_sketching = optimal_sketching

        self.reduced_rank = reduced_rank
        self._is_fitted = False
        self._lookback_len = -1

    @property
    def lookback_len(self) -> bool:
        return self._lookback_len

    @property
    def is_fitted(self) -> bool:
        return self._is_fitted

    def kernel(self, X: np.ndarray, Y: Optional[np.ndarray] = None) -> np.ndarray:
        X = X.reshape(X.shape[0], -1)
        if Y is not None:
            Y = Y.reshape(Y.shape[0], -1)
        return self._kernel(X, Y)

    def fit(self, data: np.ndarray, verbose: bool = True) -> KernelDMD:
        """
        Fits the KernelDMD model using either a randomized or a non-randomized algorithm, and either a full rank or a reduced rank algorithm,
        depending on the parameters of the model.


        .. attention::

            If ``context_len = data.shape[1] > 2``, the attribute :attr:`lookback_len` will be automatically set to ``context_len - 1``. The data will be first flattened along the trailing dimensions before computing the kernel function. The pseudo-code of this operation is

            .. code-block:: python

                X = data[:, :self.lookback_len, ...] #Y = data[:, 1:self.lookback_len + 1, ...]
                #Flatten the trailing dimensions
                X = X.reshape(data.shape[0], -1)
                self.kernel_X = self.kernel(X)

        Args:
            data (np.ndarray): Batch of context windows of shape ``(n_samples, context_len, *features_shape)``.

        Returns:
            The fitted estimator.
        """
        self._pre_fit_checks(data)
        if self.reduced_rank:
            if self.svd_solver == "randomized":
                if self.tikhonov_reg == 0.0:
                    raise ValueError(
                        "tikhonov_reg must be specified when solver is randomized."
                    )
                else:
                    U, V, spectral_bias = dual.fit_rand_reduced_rank_regression(
                        self.kernel_X,
                        self.kernel_Y,
                        self.tikhonov_reg,
                        self.rank,
                        self.n_oversamples,
                        self.optimal_sketching,
                        self.iterated_power,
                        rng_seed=self.rng_seed,
                    )
            else:
                U, V, spectral_bias = dual.fit_reduced_rank_regression(
                    self.kernel_X,
                    self.kernel_Y,
                    self.tikhonov_reg,
                    self.rank,
                    self.svd_solver,
                )
        else:
            if self.svd_solver == "randomized":
                U, V, spectral_bias = dual.fit_rand_principal_component_regression(
                    self.kernel_X,
                    self.tikhonov_reg,
                    self.rank,
                    self.n_oversamples,
                    self.iterated_power,
                    rng_seed=self.rng_seed,
                )
            else:
                U, V, spectral_bias = dual.fit_principal_component_regression(
                    self.kernel_X, self.tikhonov_reg, self.rank, self.svd_solver
                )
        self.U = U
        self.V = V
        self.rank = U.shape[1]
        self._spectral_biases = spectral_bias

        # Final Checks
        check_is_fitted(
            self,
            ["U", "V", "kernel_X", "kernel_Y", "kernel_YX", "data_fit", "lookback_len"],
        )
        self._is_fitted = True
        if verbose:
            print(
                f"Fitted {self.__class__.__name__} model. Lookback length set to {self.lookback_len}"
            )
        return self

    def risk(self, data: Optional[np.ndarray] = None) -> float:
        """Risk of the estimator on the validation ``data``.

        Args:
            data (np.ndarray or None): Batch of context windows of shape ``(n_samples, context_len, *features_shape)``. If ``None``, evaluates the risk on the training data.

        Returns:
            Risk of the estimator, see Equation 11 of :footcite:p:`Kostic2022` for more details.
        """
        if data is not None:
            check_contexts_shape(data, self.lookback_len)
            data = np.asanyarray(data)
            if data.shape[1] - 1 != self.lookback_len:
                raise ShapeError(
                    f"The  context length ({data.shape[1]}) of the validation data does not match the context length of the training data ({self.lookback_len + 1})."
                )

            X_val, Y_val = contexts_to_markov_train_states(data, self.lookback_len)
            X_train, Y_train = contexts_to_markov_train_states(
                self.data_fit, self.lookback_len
            )
            kernel_Yv = self.kernel(Y_val)
            kernel_XXv = self.kernel(X_train, X_val)
            kernel_YYv = self.kernel(Y_train, Y_val)
        else:
            kernel_Yv = self.kernel_Y
            kernel_XXv = self.kernel_X
            kernel_YYv = self.kernel_Y
        return dual.estimator_risk(
            kernel_Yv, self.kernel_Y, kernel_XXv, kernel_YYv, self.U, self.V
        )

    def predict(
        self,
        data: np.ndarray,
        t: int = 1,
        observables: Optional[Callable] = None,
    ) -> np.ndarray:
        """
        Predicts the state or, if the system is stochastic, its expected value :math:`\mathbb{E}[X_t | X_0 = X]` after ``t`` instants given the initial conditions ``X = data[:, self.lookback_len:, ...]`` being the lookback slice of ``data``.

        .. attention::

            ``data.shape[1]`` must match the lookback length ``self.lookback_len``. Otherwise, an error is raised.

        If ``observables`` are not ``None``, returns the analogue quantity for the observable instead.

        Args:
            data (numpy.ndarray): Initial conditions to predict. Array of context windows with shape ``(n_init_conditions, self.lookback_len, *self.data_fit.shape[2:])`` (see the note above).
            t (int): Number of steps in the future to predict (returns the last one).
            observables (callable or None): Callable or ``None``. If callable should map batches of states of shape ``(batch, *self.data_fit.shape[2:])`` to batches of observables ``(batch, *obs_features_shape)``.

        Returns:
           The predicted (expected) state/observable at time :math:`t`, shape ``(n_init_conditions, *obs_features_shape)``.
        """
        check_is_fitted(
            self,
            ["U", "V", "kernel_X", "kernel_Y", "kernel_YX", "data_fit", "lookback_len"],
        )

        _obs, expected_shape, X_inference, X_fit = parse_observables(
            observables, data, self.data_fit, self.lookback_len
        )

        K_Xin_X = self.kernel(X_inference, X_fit)
        return dual.predict(t, self.U, self.V, self.kernel_YX, K_Xin_X, _obs).reshape(
            expected_shape
        )

    def eig(
        self,
        eval_left_on: Optional[np.ndarray] = None,
        eval_right_on: Optional[np.ndarray] = None,
    ):
        """
        Returns the eigenvalues of the Koopman/Transfer operator and optionally evaluates left and right eigenfunctions.

        Args:
            eval_left_on (numpy.ndarray or None): Array of context windows on which the left eigenfunctions are evaluated, shape ``(n_samples, *self.data_fit.shape[1:])``.
            eval_right_on (numpy.ndarray or None): Array of context windows on which the right eigenfunctions are evaluated, shape ``(n_samples, *self.data_fit.shape[1:])``.

        Returns:
            (eigenvalues, left eigenfunctions, right eigenfunctions) - Eigenvalues of the Koopman/Transfer operator, shape ``(rank,)``. If ``eval_left_on`` or ``eval_right_on``  are not ``None``, returns the left/right eigenfunctions evaluated at ``eval_left_on``/``eval_right_on``: shape ``(n_samples, rank)``.
        """

        check_is_fitted(
            self,
            ["U", "V", "kernel_X", "kernel_Y", "kernel_YX", "lookback_len", "data_fit"],
        )
        if hasattr(self, "_eig_cache"):
            w, vl, vr = self._eig_cache
        else:
            w, vl, vr = dual.estimator_eig(
                self.U, self.V, self.kernel_X, self.kernel_YX
            )
            self._eig_cache = (w, vl, vr)

        X_fit, Y_fit = contexts_to_markov_train_states(self.data_fit, self.lookback_len)
        if eval_left_on is None and eval_right_on is None:
            # (eigenvalues,)
            return w
        elif eval_left_on is None and eval_right_on is not None:
            # (eigenvalues, right eigenfunctions)
            check_contexts_shape(
                eval_right_on, self.lookback_len, is_inference_data=True
            )
            kernel_Xin_X_or_Y = self.kernel(eval_right_on, X_fit)
            return w, dual.evaluate_eigenfunction(kernel_Xin_X_or_Y, vr)
        elif eval_left_on is not None and eval_right_on is None:
            # (eigenvalues, left eigenfunctions)
            check_contexts_shape(
                eval_left_on, self.lookback_len, is_inference_data=True
            )
            kernel_Xin_X_or_Y = self.kernel(eval_left_on, Y_fit)
            return w, dual.evaluate_eigenfunction(kernel_Xin_X_or_Y, vl)
        elif eval_left_on is not None and eval_right_on is not None:
            # (eigenvalues, left eigenfunctions, right eigenfunctions)

            check_contexts_shape(
                eval_right_on, self.lookback_len, is_inference_data=True
            )
            check_contexts_shape(
                eval_left_on, self.lookback_len, is_inference_data=True
            )

            kernel_Xin_X_or_Y_left = self.kernel(eval_left_on, Y_fit)
            kernel_Xin_X_or_Y_right = self.kernel(eval_right_on, X_fit)
            return (
                w,
                dual.evaluate_eigenfunction(kernel_Xin_X_or_Y_left, vl),
                dual.evaluate_eigenfunction(kernel_Xin_X_or_Y_right, vr),
            )

    def modes(
        self,
        data: np.ndarray,
        observables: Optional[Callable] = None,
    ) -> np.ndarray:
        """
        Computes the mode decomposition of arbitrary observables of the Koopman/Transfer operator at the states defined by ``data``.

        Informally, if :math:`(\\lambda_i, \\xi_i, \\psi_i)_{i = 1}^{r}` are eigentriplets of the Koopman/Transfer operator, for any observable :math:`f` the i-th mode of :math:`f` at :math:`x` is defined as: :math:`\\lambda_i \\langle \\xi_i, f \\rangle \\psi_i(x)`. See :footcite:t:`Kostic2022` for more details.

        Args:
            data (numpy.ndarray): Initial conditions to compute the modes on. See :func:`predict` for additional details.
            observables (callable or None): Callable or ``None``. If callable should map batches of states of shape ``(batch, *self.data_fit.shape[2:])`` to batches of observables ``(batch, *obs_features_shape)``.

        Returns:
            Modes of the system at the states defined by ``data``. Array of shape ``(self.rank, n_states, ...)``.
        """
        check_is_fitted(
            self, ["U", "V", "kernel_X", "kernel_YX", "lookback_len", "data_fit"]
        )

        _obs, expected_shape, X_inference, X_fit = parse_observables(
            observables, data, self.data_fit, self.lookback_len
        )

        if hasattr(self, "_eig_cache"):
            _, lv, rv = self._eig_cache
        else:
            _, lv, rv = dual.estimator_eig(
                self.U, self.V, self.kernel_X, self.kernel_YX
            )

        K_Xin_X = self.kernel(X_inference, X_fit)
        _gamma = dual.estimator_modes(K_Xin_X, rv, lv)

        expected_shape = (self.rank,) + expected_shape
        return np.matmul(_gamma, _obs).reshape(
            expected_shape
        )  # [rank, num_initial_conditions, num_observables]

    def svals(self):
        """Singular values of the Koopman/Transfer operator.

        Returns:
            The estimated singular values of the Koopman/Transfer operator. Array of shape `(rank,)`.
        """
        check_is_fitted(self, ["U", "V", "kernel_X", "kernel_Y"])
        return dual.svdvals(self.U, self.V, self.kernel_X, self.kernel_Y)

    def _init_kernels(self, X: np.ndarray, Y: np.ndarray):
        K_X = self.kernel(X)
        K_Y = self.kernel(Y)
        K_YX = self.kernel(Y, X)
        return K_X, K_Y, K_YX

    def _pre_fit_checks(self, data: np.ndarray) -> None:
        """Performs pre-fit checks on the training data.

        Use :func:`check_contexts_shape` to check and sanitize the input data, initialize the kernel matrices and saves the training data.

        Args:
            data (np.ndarray): Batch of context windows of shape ``(n_samples, context_len, *features_shape)``.
        """
        lookback_len = data.shape[1] - 1
        check_contexts_shape(data, lookback_len)
        data = np.asanyarray(data)
        # Save the lookback length as a private attribute of the model
        self._lookback_len = lookback_len
        X_fit, Y_fit = contexts_to_markov_train_states(data, self.lookback_len)

        self.kernel_X, self.kernel_Y, self.kernel_YX = self._init_kernels(X_fit, Y_fit)
        self.data_fit = data

        if hasattr(self, "_eig_cache"):
            del self._eig_cache
        self._check_rank(self.kernel_X.shape[0])

    def _check_rank(self, n_samples):
        if not isinstance(self.rank, int) or self.rank < 1:
            raise ValueError("rank must be a positive integer.")
        if self.rank > n_samples:
            raise ValueError(
                f"rank must be less than the number of samples ({n_samples})."
            )

    def save(self, filename):
        """Serialize the model to a file.

        Args:
            filename (path-like or file-like): Save the model to file.
        """
        pickle_save(self, filename)

    @classmethod
    def load(cls, filename):
        """Load a serialized model from a file.

        Args:
            filename (path-like or file-like): Load the model from file.

        Returns:
            KernelDMD: The loaded model.
        """
        return pickle_load(cls, filename)



from __future__ import annotations

import abc
import os
from typing import Callable, Optional, Union

import numpy as np


# Abstract base classes defining the interface to implement when extending kooplearn
class BaseModel(abc.ABC):
    @abc.abstractmethod
    def fit(self, *a, **kw):
        """Fit the model to the data. The signature of this function must be specified from the derived class. For example, in :class:`kooplearn.models.ExtendedDMD`, the signature is ``fit(self, data: np.ndarray)``, while in -TODO add AutoEncoder example."""
        pass

    @abc.abstractmethod
    def predict(
        self,
        data: np.ndarray,
        t: int = 1,
        observables: Optional[Union[Callable, np.ndarray]] = None,
    ):
        """
        Predicts the state or, if the system is stochastic, its expected value :math:`\mathbb{E}[X_t | X_0 = X]` after ``t`` instants given the initial conditions ``X = data[:, self.lookback_len:, ...]`` being the lookback slice of ``data``.

        If ``observables`` are not ``None``, returns the analogue quantity for the observable instead.

        Args:
            data (numpy.ndarray): Initial conditions to predict. Array of context windows with shape ``(n_init_conditions, context_len, *features_shape)`` whose trailing dimensions match the dimensions of the data used in :func:`fit`.
            t (int): Number of steps to predict (return the last one).
            observables (callable, numpy.ndarray or None): Callable, array of context windows of shape ``(n_init_conditions, context_len, *obs_features_shape)`` or ``None``. If array, it must be the observable evaluated at the training data. If ``None`` returns the predictions for the state.

        Returns:
           The predicted (expected) state/observable at time :math:`t`, shape ``(n_init_conditions, n_obs_features)``.
        """
        pass

    @abc.abstractmethod
    def eig(
        self,
        eval_left_on: Optional[np.ndarray] = None,
        eval_right_on: Optional[np.ndarray] = None,
    ):
        """
        Returns the eigenvalues of the Koopman/Transfer operator and optionally evaluates left and right eigenfunctions.

        Args:
            eval_left_on (numpy.ndarray or None): Array of context windows on which the left eigenfunctions are evaluated, shape ``(n_samples, context_len, *features_shape)``.
            eval_right_on (numpy.ndarray or None): Array of context windows on which the right eigenfunctions are evaluated, shape ``(n_samples, context_len, *features_shape)``.

        Returns:
            Eigenvalues of the Koopman/Transfer operator, shape ``(rank,)``. If ``eval_left_on`` or ``eval_right_on``  are not ``None``, returns the left/right eigenfunctions evaluated at ``eval_left_on``/``eval_right_on``: shape ``(n_samples, rank)``.
        """
        pass

    @abc.abstractmethod
    def save(self, path: os.PathLike):
        pass

    @classmethod
    @abc.abstractmethod
    def load(path: os.PathLike):
        pass

    @abc.abstractmethod
    def modes(
        self,
        data: np.ndarray,
        observables: Optional[Union[Callable, np.ndarray]] = None,
    ):
        """
        Computes the mode decomposition of arbitrary observables of the Koopman/Transfer operator at the states defined by ``data``.

        Informally, if :math:`(\\lambda_i, \\xi_i, \\psi_i)_{i = 1}^{r}` are eigentriplets of the Koopman/Transfer operator, for any observable :math:`f` the i-th mode of :math:`f` at :math:`x` is defined as: :math:`\\lambda_i \\langle \\xi_i, f \\rangle \\psi_i(x)`. See :footcite:t:`Kostic2022` for more details.

        Args:
            data (numpy.ndarray): Initial conditions to compute the modes on. See :func:`predict` for additional details.
            observables (callable, numpy.ndarray or None): Callable, array of context windows of shape ``(n_samples, context_len, *obs_features_shape)`` or ``None``. If array, it must be the desired observable evaluated on the *lookforward slice* of the training data. If ``None`` returns the predictions for the state.
        Returns:
            Modes of the system at the states defined by ``data``. Array of shape ``(rank, n_samples, ...)``.
        """
        pass

    @property
    @abc.abstractmethod
    def is_fitted(self) -> bool:
        """Check if the model is fitted.

        Returns:
            Returns ``True`` if the model is fitted, ``False`` otherwise.
        """
        pass

    @property
    @abc.abstractmethod
    def lookback_len(self) -> int:
        """Length of the lookback window associated to the contexts. Upon fitting, the dimension of the lookforward window will be inferred from the context window length and this attribute. Moreover, shape checks against this attribute will be performed on the data passed to :func:`fit`, :func:`predict`, :func:`eig` and :func:`modes`.

        Returns:
            Length of the lookback window associated to the contexts.
        """
        pass


class FeatureMap(abc.ABC):
    """Abstract base class for feature maps. The :func:`__call__` method must accept a batch of data points of shape ``(n_samples, *features_shape)`` and return a batch of features of shape ``(n_samples, out_features)``.

    .. caution::

        As described in :ref:`kooplearn's data paradigm <kooplearn_data_paradigm>`, the inputs passed to the methods of :class:`BaseModel` will be in the form of batches of context windows. :class:`FeatureMap`, instances are a notable departure from this paradigm, and we expect a feature map to be called on batches of data points, not context windows. The context windows are parsed internally by :class:`BaseModel` and passed to the feature map as batches of appropriate data points.

        This behaviour is designed to facilitate the reuse of feature maps across different models, and possibly even outside of kooplearn.
    """

    @abc.abstractmethod
    def __call__(self, data: np.ndarray) -> np.ndarray:
        pass


class TrainableFeatureMap(FeatureMap):
    @abc.abstractmethod
    def fit(self, *a, **kw) -> None:
        """Fit the feature map to the data."""
        pass

    @property
    @abc.abstractmethod
    def is_fitted(self) -> bool:
        pass

    @property
    @abc.abstractmethod
    def lookback_len(self) -> int:
        """Length of the lookback window associated to the contexts. Upon fitting, the dimension of the lookforward window will be inferred from the context window length.

        Returns:
            Length of the lookback window associated to the contexts.
        """
        pass



from typing import Optional, Union

import torch

from kooplearn._src.utils import ShapeError


# A bit of code copy paste but it's ok for now
def encode_contexts(contexts_batch: torch.Tensor, encoder: torch.nn.Module):
    # Caution: this method is designed only for internal calling.
    context_len = contexts_batch.shape[1]
    Z = []
    for i in range(context_len):  # Inefficient but working
        X = contexts_batch[:, i, ...]
        Z.append(encoder(X))
    Z = torch.stack(Z, dim=1)
    latent_dim = Z.shape[2:]
    assert (
        len(latent_dim) == 1
    ), "The encoder output must be a 1-dimensional tensor (i.e. a vector)."
    return Z


def decode_contexts(encoded_contexts_batch: torch.Tensor, decoder: torch.nn.Module):
    # Caution: this method is designed only for internal calling.
    context_len = encoded_contexts_batch.shape[1]
    assert (
        len(encoded_contexts_batch.shape[2:]) == 1
    ), "The decoder input must be a 1-dimensional tensor (i.e. a vector)."
    Z = []
    for i in range(context_len):  # Inefficient but working
        X = encoded_contexts_batch[:, i, ...]
        Z.append(decoder(X))
    Z = torch.stack(Z, dim=1)
    return Z


def evolve_contexts(
    encoded_contexts_batch: torch.Tensor,
    lookback_len: int,
    forward_operator: Union[torch.nn.Parameter, torch.Tensor],
    backward_operator: Optional[Union[torch.nn.Parameter, torch.Tensor]] = None,
):
    # Caution: this method is designed only for internal calling.
    context_len = encoded_contexts_batch.shape[1]
    X_init = encoded_contexts_batch[:, lookback_len - 1, ...]  # Initial condition
    if X_init.ndim != 2:
        raise ShapeError(
            f"The encoder network must return a 1D vector for each snapshot, while a shape {X_init.shape[1:]} tensor was received."
        )
    evolved_contexts_batch = torch.zeros_like(encoded_contexts_batch)

    for ctx_idx in range(context_len):  # Not efficient but working
        exp = ctx_idx - lookback_len + 1
        if exp < 0:
            if backward_operator is None:
                raise ValueError(
                    "No backward operator provided, but lookback length greater than 1."
                )
            else:
                pwd_operator = torch.matrix_power(backward_operator, -exp)
                Z = torch.mm(pwd_operator, X_init.T).T
        else:
            pwd_operator = torch.matrix_power(forward_operator, exp)
            Z = torch.mm(pwd_operator, X_init.T).T
        evolved_contexts_batch[:, ctx_idx, ...] = Z
    return evolved_contexts_batch  # [batch_size, context_len, latent_dim]


def evolve_batch(
    batch: torch.Tensor,
    t: int,
    encoder: torch.nn.Module,
    decoder: torch.nn.Module,
    forward_operator: Union[torch.nn.Parameter, torch.Tensor],
):
    Z = encoder(batch)
    assert (
        len(Z.shape[1:]) == 1
    ), "The encoder output must be a 1-dimensional tensor (i.e. a vector)."
    evolution_operator = torch.matrix_power(forward_operator, t)
    Z_evolved = torch.mm(evolution_operator, Z.T).T
    batch_evolved = decoder(Z_evolved)
    assert batch.shape == batch_evolved.shape
    return batch_evolved


def consistency_loss(
    forward_operator: Union[torch.nn.Parameter, torch.Tensor],
    backward_operator: Union[torch.nn.Parameter, torch.Tensor],
):
    assert (
        forward_operator.shape == backward_operator.shape
    ), "The forward and backward operators must have the same shape."
    terms = []
    dim = forward_operator.shape[0]
    for k in range(1, dim + 1):
        fb = torch.mm(forward_operator[:k, :], backward_operator[:, :k]) - torch.eye(
            k, device=forward_operator.device
        )
        bf = torch.mm(backward_operator[:k, :], forward_operator[:, :k]) - torch.eye(
            k, device=forward_operator.device
        )
        terms.append(
            torch.linalg.matrix_norm(fb) ** 2 + torch.linalg.matrix_norm(bf) ** 2
        )
    return 0.5 * torch.mean(torch.stack(terms))



from pathlib import Path
from shutil import rmtree

import numpy as np
import pytest
from scipy.stats import special_ortho_group
from sklearn.gaussian_process.kernels import RBF, DotProduct, Matern

from kooplearn.data import traj_to_contexts
from kooplearn.datasets.stochastic import LinearModel
from kooplearn.models import KernelDMD

TRUE_RANK = 5
DIM = 20
NUM_SAMPLES = 50


def make_linear_system():
    eigs = 9 * np.logspace(-3, -1, TRUE_RANK)
    # print("Eigenvalues:")
    # for ev in eigs:
    #     print(f"{ev:.1e} \t", end='')
    eigs = np.concatenate([eigs, np.zeros(DIM - TRUE_RANK)])
    Q = special_ortho_group(DIM, 0).rvs(1)
    A = np.linalg.multi_dot([Q, np.diag(eigs), Q.T])

    # Consistency-check
    assert np.allclose(np.sort(np.linalg.eigvalsh(A)), np.sort(eigs))
    return LinearModel(A, noise=1e-5, rng_seed=0)


@pytest.mark.parametrize("kernel", [DotProduct(), RBF(), Matern()])
@pytest.mark.parametrize("reduced_rank", [True, False])
@pytest.mark.parametrize("rank", [TRUE_RANK, TRUE_RANK - 2, TRUE_RANK + 10])
@pytest.mark.parametrize("solver", ["full", "arnoldi", "wrong"])
@pytest.mark.parametrize("tikhonov_reg", [None, 0.0, 1e-5])
@pytest.mark.parametrize("observables", [None, lambda x: x[:, 0]])
@pytest.mark.parametrize("lookback_len", [1, 2, 3])
def test_KernelDMD_fit_predict_eig_modes_save_load(
    kernel, reduced_rank, rank, solver, tikhonov_reg, observables, lookback_len
):
    dataset = make_linear_system()
    _Z = dataset.sample(np.zeros(DIM), NUM_SAMPLES)
    data = traj_to_contexts(_Z, lookback_len + 1)
    if solver not in ["full", "arnoldi"]:
        with pytest.raises(ValueError):
            model = KernelDMD(
                kernel=kernel,
                reduced_rank=reduced_rank,
                rank=rank,
                tikhonov_reg=tikhonov_reg,
                svd_solver=solver,
            )
    else:
        model = KernelDMD(
            kernel=kernel,
            reduced_rank=reduced_rank,
            rank=rank,
            tikhonov_reg=tikhonov_reg,
            svd_solver=solver,
        )

        assert model.is_fitted is False
        model.fit(data)
        test_data = data[:, :lookback_len, ...]
        assert model.is_fitted is True
        if observables is None:
            X_pred = model.predict(test_data, observables=observables)
            assert X_pred.shape == (data.shape[0],) + data.shape[2:]
            modes = model.modes(test_data, observables=observables)
            assert modes.shape[1:] == (data.shape[0],) + data.shape[2:]
        else:
            Y = data[:, -1, ...]
            _dummy_vec = observables(Y)
            if _dummy_vec.ndim == 1:
                _dummy_vec = _dummy_vec[:, None]
            X_pred = model.predict(test_data, observables=observables)
            assert X_pred.shape == _dummy_vec.shape
            modes = model.modes(test_data, observables=observables)
            _target_shape = _dummy_vec.shape
            assert modes.shape[1:] == _target_shape

        vals, lv, rv = model.eig(eval_left_on=test_data, eval_right_on=test_data)
        assert vals.shape[0] <= rank
        assert vals.ndim == 1
        tmp_path = Path(__file__).parent / f"tmp/model.bin"
        model.save(tmp_path)
        restored_model = KernelDMD.load(tmp_path)

        assert np.allclose(model.kernel_X, restored_model.kernel_X)
        assert np.allclose(model.kernel_Y, restored_model.kernel_Y)
        assert np.allclose(model.kernel_YX, restored_model.kernel_YX)
        assert np.allclose(model.data_fit, restored_model.data_fit)
        assert np.allclose(model.lookback_len, model.lookback_len)
        rmtree(Path(__file__).parent / "tmp/")



from typing import NamedTuple

import numpy as np
import pytest

from kooplearn._src.operator_regression import dual, primal
from kooplearn.datasets import Mock


class EigenDecomposition(NamedTuple):
    values: np.ndarray
    left: np.ndarray
    right: np.ndarray


def _allclose(a, b):
    return np.allclose(a, b, rtol=1e-3, atol=1e-5)


def _compare_up_to_sign(a: np.ndarray, b: np.ndarray) -> bool:
    return _allclose(np.abs(a), np.abs(b))


def _compare_evd(evd_1: EigenDecomposition, evd_2: EigenDecomposition) -> bool:
    _ev1 = np.copy(evd_1.values)
    _ev2 = np.copy(evd_2.values)
    evd_1_sort = np.argsort(_ev1, kind="stable")
    evd_2_sort = np.argsort(_ev2, kind="stable")

    assert _allclose(_ev1[evd_1_sort], _ev2[evd_2_sort])
    assert _compare_up_to_sign(evd_1.left[:, evd_1_sort], evd_2.left[:, evd_2_sort])
    assert _compare_up_to_sign(evd_1.right[:, evd_1_sort], evd_2.right[:, evd_2_sort])
    return True


@pytest.mark.parametrize("tikhonov_reg", [1e-3])
@pytest.mark.parametrize("svd_solver", ["full", "arnoldi"])
@pytest.mark.parametrize("dt", [1, 2, 3])
def test_reduced_rank_tikhonov_primal_dual_consistency(dt, svd_solver, tikhonov_reg):
    num_features = 10
    num_test_pts = 100
    rank = 4

    dataset = Mock(num_features=num_features, rng_seed=42)
    _Z = dataset.sample(None, num_test_pts)
    X, Y = _Z[:-1], _Z[1:]

    rng = np.random.default_rng(42)
    X_test = rng.random((num_test_pts, num_features))

    rdim = np.true_divide(1, X.shape[0])
    K_X = X @ (X.T)
    C_X = rdim * ((X.T) @ X)

    K_Y = Y @ (Y.T)
    C_Y = rdim * ((Y.T) @ Y)

    K_YX = Y @ (X.T)
    C_XY = rdim * ((X.T) @ Y)

    K_testX = X_test @ (X.T)

    # Dual
    U, V, _ = dual.fit_reduced_rank_regression(
        K_X, K_Y, tikhonov_reg, rank, svd_solver=svd_solver
    )
    dual_predict = dual.predict(dt, U, V, K_YX, K_testX, Y)
    dual_eig, dual_lv, dual_rv = dual.estimator_eig(U, V, K_X, K_YX)
    dual_modes = dual.estimator_modes(X_test @ (X.T), dual_rv, dual_lv)

    evd_dual = EigenDecomposition(
        dual_eig,
        dual.evaluate_eigenfunction(X_test @ (Y.T), dual_lv),
        dual.evaluate_eigenfunction(X_test @ (X.T), dual_rv),
    )

    assert dual_predict.shape == (num_test_pts, num_features)

    # Primal
    U_primal = primal.fit_reduced_rank_regression(
        C_X, C_XY, tikhonov_reg, rank, svd_solver=svd_solver
    )

    primal_predict = primal.predict(dt, U_primal, C_XY, X_test, X, Y)
    primal_eig, primal_lv, primal_rv = primal.estimator_eig(U_primal, C_XY)
    primal_modes = primal.estimator_modes(U_primal, C_XY, X, X_test)

    evd_primal = EigenDecomposition(
        primal_eig,
        primal.evaluate_eigenfunction(X_test, primal_lv),
        primal.evaluate_eigenfunction(X_test, primal_rv),
    )

    assert dual_predict.shape == (num_test_pts, num_features)
    risk_primal = primal.estimator_risk(C_X, C_Y, C_XY, C_XY, U_primal)
    risk_dual = dual.estimator_risk(K_Y, K_Y, K_X, K_Y, U, V)
    assert _allclose(risk_primal, risk_dual)
    assert _allclose(primal_modes, dual_modes)
    assert _allclose(primal_predict, dual_predict)
    assert _compare_evd(evd_primal, evd_dual)


@pytest.mark.parametrize("tikhonov_reg", [0.0, 1e-3])
@pytest.mark.parametrize("rank", [5, 7])
@pytest.mark.parametrize("svd_solver", ["full", "arnoldi"])
@pytest.mark.parametrize("dt", [1, 2, 3])
def test_tikhonov_primal_dual_consistency(dt, svd_solver, rank, tikhonov_reg):
    num_features = 10
    num_test_pts = 100

    dataset = Mock(num_features=num_features, rng_seed=42)
    _Z = dataset.sample(None, num_test_pts)
    X, Y = _Z[:-1], _Z[1:]

    rng = np.random.default_rng(42)
    X_test = rng.random((num_test_pts, num_features))

    rdim = np.true_divide(1, X.shape[0])
    K_X = X @ (X.T)
    C_X = rdim * ((X.T) @ X)

    K_YX = Y @ (X.T)
    C_XY = rdim * ((X.T) @ Y)

    K_Y = Y @ (Y.T)
    C_Y = rdim * ((Y.T) @ Y)

    K_testX = X_test @ (X.T)

    # Dual
    U, V, _ = dual.fit_principal_component_regression(
        K_X, tikhonov_reg, rank=rank, svd_solver=svd_solver
    )

    dual_predict = dual.predict(dt, U, V, K_YX, K_testX, Y)

    dual_eig, dual_lv, dual_rv = dual.estimator_eig(U, V, K_X, K_YX)
    dual_modes = dual.estimator_modes(X_test @ (X.T), dual_rv, dual_lv)

    evd_dual = EigenDecomposition(
        dual_eig,
        dual.evaluate_eigenfunction(X_test @ (Y.T), dual_lv),
        dual.evaluate_eigenfunction(X_test @ (X.T), dual_rv),
    )

    assert dual_predict.shape == (num_test_pts, num_features)

    # Primal
    U_primal = primal.fit_principal_component_regression(
        C_X, tikhonov_reg, rank=rank, svd_solver=svd_solver
    )

    primal_predict = primal.predict(dt, U_primal, C_XY, X_test, X, Y)
    primal_eig, primal_lv, primal_rv = primal.estimator_eig(U_primal, C_XY)
    primal_modes = primal.estimator_modes(U_primal, C_XY, X, X_test)

    evd_primal = EigenDecomposition(
        primal_eig,
        primal.evaluate_eigenfunction(X_test, primal_lv),
        primal.evaluate_eigenfunction(X_test, primal_rv),
    )

    assert primal_predict.shape == (num_test_pts, num_features)
    risk_primal = primal.estimator_risk(C_X, C_Y, C_XY, C_XY, U_primal)
    risk_dual = dual.estimator_risk(K_Y, K_Y, K_X, K_Y, U, V)
    assert _allclose(risk_primal, risk_dual)
    assert _allclose(primal_predict, dual_predict)
    if rank is not None:
        assert _allclose(primal_modes, dual_modes)
        assert _compare_evd(evd_primal, evd_dual)


@pytest.mark.skip()
@pytest.mark.parametrize("dt", [1, 5, 10])
def test_rand_reduced_rank(dt):
    num_features = 50
    num_test_pts = 200
    rank = 3
    tikhonov_reg = 1e-3
    n_oversamples = 20
    iterated_power = 2

    dataset = Mock(num_features=num_features, rng_seed=42)
    _Z = dataset.sample(None, num_test_pts)
    X, Y = _Z[:-1], _Z[1:]

    rng = np.random.default_rng(42)
    X_test = rng.random((num_test_pts, num_features))

    rdim = np.true_divide(1, X.shape[0])
    K_X = X @ (X.T)
    C_X = rdim * ((X.T) @ X)

    K_Y = Y @ (Y.T)

    K_YX = Y @ (X.T)
    C_XY = rdim * ((X.T) @ Y)

    K_testX = X_test @ (X.T)

    # Dual
    U, V, _ = dual.fit_rand_reduced_rank_regression(
        K_X, K_Y, rank, tikhonov_reg, n_oversamples, False, iterated_power
    )

    dual_predict = dual.low_rank_predict(dt, U, V, K_YX, K_testX, Y)
    dual_eig, _, _ = dual.low_rank_eig(U, V, K_X, K_Y, K_YX)

    assert dual_predict.shape == (num_test_pts, num_features)

    # Primal
    U = primal.fit_rand_reduced_rank_regression(
        C_X, C_XY, rank, tikhonov_reg, n_oversamples, iterated_power
    )

    primal_predict = primal.low_rank_predict(dt, U, C_XY, X_test, X, Y)
    primal_eig, _ = primal.low_rank_eig(U, C_XY)

    assert dual_predict.shape == (num_test_pts, num_features)
    assert _allclose(primal_predict, dual_predict)
    assert _allclose(np.sort(dual_eig.real), np.sort(primal_eig.real))
    assert _allclose(np.sort(dual_eig.imag), np.sort(primal_eig.imag))


@pytest.mark.skip()
@pytest.mark.parametrize("dt", [1, 5, 10])
def test_rand_reduced_rank_primal(dt):
    num_features = 50
    num_test_pts = 200
    rank = 5
    tikhonov_reg = 1e-3
    n_oversamples = 20
    iterated_power = 3

    dataset = Mock(num_features=num_features, rng_seed=42)
    _Z = dataset.sample(None, num_test_pts)
    X, Y = _Z[:-1], _Z[1:]

    rng = np.random.default_rng(42)
    X_test = rng.random((num_test_pts, num_features))

    rdim = np.true_divide(1, X.shape[0])
    C_X = rdim * ((X.T) @ X)
    C_XY = rdim * ((X.T) @ Y)

    # Primal
    U = primal.fit_reduced_rank_regression(C_X, C_XY, rank, tikhonov_reg)

    predict = primal.low_rank_predict(dt, U, C_XY, X_test, X, Y)
    eig, _ = primal.low_rank_eig(U, C_XY)

    # Rand
    U_rand = primal.fit_rand_reduced_rank_regression(
        C_X, C_XY, rank, tikhonov_reg, n_oversamples, iterated_power
    )

    rand_predict = primal.low_rank_predict(dt, U_rand, C_XY, X_test, X, Y)
    rand_eig, _ = primal.low_rank_eig(U_rand, C_XY)

    diff = np.ravel(rand_predict - predict)
    mean = 0.5 * np.ravel(rand_predict + predict)
    print(np.linalg.norm(diff, ord=np.inf) / np.linalg.norm(mean, ord=np.inf))

    assert predict.shape == (num_test_pts, num_features)
    assert rand_predict.shape == (num_test_pts, num_features)

    assert _allclose(predict, rand_predict)
    assert _allclose(np.sort(rand_eig.real), np.sort(eig.real))
    assert _allclose(np.sort(rand_eig.imag), np.sort(eig.imag))


@pytest.mark.skip()
@pytest.mark.parametrize("dt", [1, 5, 10])
def test_rand_reduced_rank_dual(dt):
    num_features = 50
    num_test_pts = 200
    rank = 10
    tikhonov_reg = 1e-3
    n_oversamples = 10
    iterated_power = 2

    dataset = Mock(num_features=num_features, rng_seed=42)
    _Z = dataset.sample(None, num_test_pts)
    X, Y = _Z[:-1], _Z[1:]

    rng = np.random.default_rng(42)
    X_test = rng.random((num_test_pts, num_features))

    K_X = X @ (X.T)
    K_Y = Y @ (Y.T)
    K_YX = Y @ (X.T)
    K_testX = X_test @ (X.T)

    # Dual
    U, V, _ = dual.fit_reduced_rank_regression(K_X, K_Y, rank, tikhonov_reg)

    predict = dual.low_rank_predict(dt, U, V, K_YX, K_testX, Y)
    eig, _, _ = dual.low_rank_eig(U, V, K_X, K_Y, K_YX)

    # Rand
    U_rand, V_rand, _ = dual.fit_rand_reduced_rank_regression(
        K_X, K_Y, rank, tikhonov_reg, n_oversamples, False, iterated_power
    )

    rand_predict = dual.low_rank_predict(dt, U_rand, V_rand, K_YX, K_testX, Y)
    rand_eig, _, _ = dual.low_rank_eig(U_rand, V_rand, K_X, K_Y, K_YX)

    assert predict.shape == (num_test_pts, num_features)
    assert rand_predict.shape == (num_test_pts, num_features)

    assert _allclose(predict, rand_predict)
    assert _allclose(np.sort(rand_eig.real), np.sort(eig.real))
    assert _allclose(np.sort(rand_eig.imag), np.sort(eig.imag))



import numpy as np
import pytest

from kooplearn._src.operator_regression import dual, primal
from kooplearn.datasets import Mock


def _primal_right_normalization(right_vectors: np.ndarray):
    _norms = np.sum(right_vectors.conj() * right_vectors, axis=0)
    return np.allclose(_norms, np.ones(_norms.shape[0]))


def _primal_eigenvalue_equation(
    eigenvalues: np.ndarray,
    left_vectors: np.ndarray,
    right_vectors: np.ndarray,
    estimator: np.ndarray,
):
    reconstruction = np.linalg.multi_dot(
        [right_vectors, np.diag(eigenvalues), left_vectors.T]
    )
    return np.allclose(estimator, reconstruction)


def _primal_biortogonality(left_vectors: np.ndarray, right_vectors: np.ndarray):
    return np.allclose((left_vectors.T) @ right_vectors, np.eye(left_vectors.shape[1]))


@pytest.mark.parametrize("tikhonov_reg", [1e-3])
@pytest.mark.parametrize("svd_solver", ["full", "arnoldi"])
def test_primal_eig_decomposition(tikhonov_reg, svd_solver):
    num_features = 20
    num_test_pts = 200
    rank = 5

    dataset = Mock(num_features=num_features, rng_seed=42)
    _Z = dataset.sample(None, num_test_pts)
    X, Y = _Z[:-1], _Z[1:]

    rdim = np.true_divide(1, X.shape[0])
    C_X = rdim * ((X.T) @ X)
    C_XY = rdim * ((X.T) @ Y)

    U = primal.fit_principal_component_regression(
        C_X, tikhonov_reg, rank=rank, svd_solver=svd_solver
    )
    eig, lv, rv = primal.estimator_eig(U, C_XY)
    estimator = np.linalg.multi_dot([U, U.T, C_XY])

    assert _primal_right_normalization(rv)
    assert _primal_eigenvalue_equation(eig, lv, rv, estimator)
    assert _primal_biortogonality(lv, rv)


def _dual_right_normalization(right_vectors: np.ndarray, K_X: np.ndarray):
    r_dim = (K_X.shape[0]) ** (-1)
    _norms = np.sum(right_vectors.conj() * (r_dim * K_X @ right_vectors), axis=0)
    return np.allclose(_norms, np.ones(_norms.shape[0]))


def _dual_biortogonality(
    left_vectors: np.ndarray, right_vectors: np.ndarray, K_YX: np.ndarray
):
    r_dim = (K_YX.shape[0]) ** (-1)
    return np.allclose(
        np.linalg.multi_dot([left_vectors.T, r_dim * K_YX, right_vectors]),
        np.eye(left_vectors.shape[1]),
    )


@pytest.mark.parametrize("tikhonov_reg", [1e-3])
@pytest.mark.parametrize("svd_solver", ["full", "arnoldi"])
def test_dual_eig_decomposition(tikhonov_reg, svd_solver):
    num_features = 20
    num_test_pts = 200
    rank = 5

    dataset = Mock(num_features=num_features, rng_seed=42)
    _Z = dataset.sample(None, num_test_pts)
    X, Y = _Z[:-1], _Z[1:]

    K_X = X @ (X.T)
    K_YX = Y @ (X.T)

    # Dual
    U, V, _ = dual.fit_principal_component_regression(
        K_X, tikhonov_reg, rank=rank, svd_solver=svd_solver
    )
    eig, lv, rv = dual.estimator_eig(U, V, K_X, K_YX)

    assert _dual_right_normalization(rv, K_X)
    assert _dual_biortogonality(lv, rv, K_YX)



import abc
from typing import NamedTuple

import numpy as np
from tqdm import tqdm


class LinalgDecomposition(NamedTuple):
    values: np.ndarray
    x: np.ndarray
    functions: np.ndarray


# General utility classes
class DataGenerator(abc.ABC):
    @abc.abstractmethod
    def sample(self, X0: np.ndarray, T: int = 1):
        pass


class DiscreteTimeDynamics(DataGenerator):
    def sample(self, X0: np.ndarray, T: int = 1, show_progress: bool = False):
        X0 = np.asarray(X0)
        if X0.ndim == 0:
            X0 = X0[None]

        memory = np.zeros((T + 1,) + X0.shape)
        memory[0] = X0
        if show_progress:
            _iter = tqdm(range(T), desc="Generating data", unit="step", leave=False)
        else:
            _iter = range(T)

        for t in _iter:
            memory[t + 1] = self._step(memory[t])
        return memory

    @abc.abstractmethod
    def _step(self, X: np.ndarray):
        pass
